/************************************/
/*	OS Assignment 3
/*	By: Simon Foucher
/*	260 223 197
/************************************/


/************************************/
/*	includes and definitions				
/************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#define TRUE 1
#define FALSE 0
typedef int BOOL;


/************************************/
/*	Function declarations				
/************************************/
void logo(void);
void TerminateProcesses();
int interpreter(char commandline[], int Running);
int GetRessourceIndex(char userInput[], int PIDRunning);
int WriteToFile(char userInput[], int Running);
void FreeResourceIndex(int INDEX);
int closeINDEX(char userInput[], int Running);
int deleteFile(char userInput[], int Running);


/************************************/
/*	Global Variables & datatypes				
/************************************/
int quanta = 1;	                            // We'll give the option to the user to set quanta
                                            // Since CPU interrupted by shell prompt every CPU cycle, 

char *FREERAM[100];							//Pointer to RAM array
int FRAMESIZE = 10;                          // Default frame size value
char *shellMemoryVar[100];
char *shellMemoryData[100];
char userPrompt[40] = ".$ ";
BOOL verbose = FALSE;

FILE *FAT;
FILE *FILES;
int FileHandles[5][2];
int BUFFER = 75;							// Memory reserved buffer size when opening a file on HDD


int concurrent = 10;
int readyHead;								// Pointer to the head of the Ready Q
int readyTail;								// Tail pointer of the ready Q 
int readyQ[10];     		                // We'll use an array for the ready queue
int terminateQ[10];
int loadingQ[50];							// Loading Q
int loadingHead;							// Head pointer to the loading Q
int loadingTail;

// The waqiting Q has 5 seperate lines, eache for a file handle
// Each subQ has it's own set of Q pointers	
int waitingQ[5][50];							// Loading Q

int waitingHead[5];							// Head pointer to the loading Q
int waitingTail[5];	

// This global variable is for the process to communicate which index it is waiting for
int INDEXwait;
// Used to notify the world that file handles got freed (1=freed,0 = nothing)
int clearedHandles[5];


struct page_rec
{
        int inRAM;							// boolean true if page in RAM, false if page in backing store
        int wasModified; 					// will not be used in this assignment
        int baseAddress; 					// the index number of the RAM array cell of the frame
};

struct PCBdef								// Process Control Block Data Structure definition
	{
		char filename[50]; 					// Stores the name of the script loaded into memory
		int start; 						    // Start address loaded into RAM
		int length; 						// Number of instructions
		int current; 						// Current instruction being executed
		char *memory[20]; 					// List of Shell Memory variables and values
		struct page_rec pageTable[100];		// max limit of 100 pages
		FILE *BACKINGSTORE;					// Backing store file pointer
		long int MAXPAGES;					// total number of created pages for program
		struct PCB *next; 					// Next node in the linked-list
	};


struct cpu
	{
	int quanta; int returnCode;
	int pc;
	char *memory[20];						// Since the PCBs only have 20 memory cells
	};


struct bs_rec								// Backing store
{
        char command1[100];
        char command2[100];
        char command3[100];
        char command4[100];
};


struct OSFAT
   {
       char filename[20]; // file names are at most 19 characters with /0
       int address; // an offset value from the start of the simulated disk
   };






struct PCBdef PCB[10];				// Creates 9 PCBs (max threads that can execute at the same time)



/************************************/
/*	Functions						
/************************************/



/*******************************************************************************/
/*	New/Modified functions since shell.c are clarly identified.
/*	All other functions are similar to old version
/*******************************************************************************/

BOOL displayMemVar(char commandline[])
{
    int i,j;
    BOOL found = FALSE;
    char variable[50];

    // variable names is the second word

    for(i=3,j=0;i<200 && (commandline[i]!='\0' || commandline[i]!='\0');i++)
      if (commandline[i]!=' ')
      {
          variable[j] = commandline[i];
          j++;
      }

    variable[j]='\0';

    // locate variable in shell memory

    for(i=0;i<100 && shellMemoryVar[i]!=NULL;i++)
    {
        if (strcmp(shellMemoryVar[i],variable)==0)
        {
            found = TRUE;
            printf("%s=%s\n",variable,shellMemoryData[i]);
            break;
        }
    }

    if (!found) {printf("Not found\n"); return FALSE;}

    return TRUE;
}

void stripCRLF(char array[])
{
    int i;

    for(i=0;i<500 && array[i]!='\0';i++)
      if (array[i]=='\r' || array[i]=='\n') array[i]='\0';
}


/*******************************************************************************/
/*	Function setMemVar: Stores variable and value
/*	
/*	Input: 
/*      The command (from RAM or SHELL)
/*      Running: a pointer to memory. If Running = 15; command came from SHELL
/*
/*  Additions:
/*      - Records the variable used by a script to its PCB
/*      - The function blocks scripts form overwrighting variables that are
/*          not part of their PCB (either created by shell or another script
/*      - If a shell command overwrights a variable reated by a script,
/*          a warning is issued but the variable get's overwritten
/*******************************************************************************/





setMemVar(char commandline[],int Running)
{
    char variable[100];
    char data[100];
    char userInput[1];
    int i,j,k;
    BOOL outOfMemory = TRUE;
    int script = TRUE;                                          // Assube that we are interpreting a script be default
    int allowed = TRUE;                                         // To make sure that processes don't overwrite other processes's memory          
    if (Running==-1)                                            // If the command came from the shell, Running =15                    
        script = FALSE;                                         // Not scripting, use PCB memory

    // skip white spaces
    for(i=3;i<200 && commandline[i]!='\0' && commandline[i]==' ';i++);

    // copy variable
    for (j=0;i<200 && commandline[i]!='\0' && commandline[i]!=' ';i++,j++)
        variable[j]=commandline[i];

    variable[j]='\0';

    // skip white spaces
    for(;i<200 && commandline[i]!='\0' && commandline[i]==' ';i++);

    // copy data
    for (j=0;i<200 && commandline[i]!='\0' && commandline[i]!=' ';i++,j++)
        data[j]=commandline[i];

    data[j]='\0';


    // Store it into the shell's memory, overwritting wit user approval it if it already exists

    for(i=0;i<100;i++)
    {
        if (shellMemoryVar[i]==NULL) // not in memory so insert and exit
        {
            allowed = TRUE;
            shellMemoryVar[i]  = strdup(variable);
            shellMemoryData[i] = strdup(data);
            outOfMemory = FALSE;
            break;
        }

        else if (strcmp(shellMemoryVar[i],variable)==0)                     // If the variable already exists in Shell memory
        {
           if(script)                                                       // If scripting, make sure we don't overwrite other processes's valriables
                {
                    allowed = FALSE;                                        // Not allowd until proven that variable is in script memory
                            for(j=0;(j<=20 && PCB[Running].memory[j] != NULL);j++)// Scanning internal memory for variable name
                            {                                                       
                                if(strcmp(PCB[Running].memory[j], variable)==0)   // If it's part of the script's memory
                                    {allowed = TRUE;break;}                       // Allow overwright and break
                            }

                    if(!allowed)                                            // Warn user of illegal operation
                      {
                        printf("\nWARNING!!!\nThe script is trying to overwright the variable '%s',\n",variable);
                        printf("which was created by another script or the shell.\n\n");
                        printf("Permission denied and instruction ignored!\n\n");
                      }
                }// End of if scripting

            else if(!script)                                                // If we are not scripting
                {
                    allowed = TRUE;                                         // Allowed to overwright                   
                    for(k=0;k<=9;k++)                                       // Scan all PCBs
                      {
                        for(j=0;(j<=20 && PCB[k].memory[j] != NULL);j++)    // Scan every non empty memory Cells of PCBs
                            {
                              if(strcmp(PCB[k].memory[j], variable)==0)     // If the variable is found in a PCB's memory
                                {                                           // Warn user 
                                    printf("WARNING!!!\nThe variable %s was created by another script.\n",variable);
                                    printf("Overwriting it's value will permanently destroy it's previous value\n");
                                    printf("without warning the script who created it. This might cause the process\n");
                                    printf("to behave abnormally or crash.\nAlso note that the variable will");
                                    printf("be deleted from Shell memory\nwhen that script ends!!\n\n");
                                    break;
                                }// End of displaying warning
                            }// End of scanning every non empty memory cells
                      }// End of scanning every PCBs
                }// End of if !scripting

            if(allowed)                                                     // So if we are allowed to overwright
                {                        
                    free(shellMemoryVar[i]);
                    free(shellMemoryData[i]);

                    shellMemoryVar[i]  = strdup(variable);
                    shellMemoryData[i] = strdup(data);

                    outOfMemory = FALSE;
                    break;
                }
        }// End of if variable already exists
    }

                       
    if(outOfMemory == FALSE && script==TRUE && allowed == TRUE)             // If we are scripting and variable got stored in SHELL memory
        {           
            for(j=0;j<20;j++)                                               // Scan PCB memory cells            
                {
                     if (PCB[Running].memory[j]==NULL)                      // Variable not in memory so insert and exit
                        {
                            PCB[Running].memory[j]  = strdup(variable);
                            break;
                        }
                      else if (strcmp(PCB[Running].memory[j],variable)==0)  // Variable already exists in PCB memory, break
                        break;

                }// End of scanning PCB memory cells
        }// End of out of !outOfmemory & scripting

    // if it was not added to the shell the outOfMemory still TRUE

    if (outOfMemory)
        printf("%s=%s not inserted.  Out of shell memory!\n",variable,data);
}



void changePrompt(char commandline[])
{
    int i,j;
    // variable in the second word

    for(i=6,j=0;i<200 && commandline[i]!='\0';i++)
      if (commandline[i]!=' ')
      {
          userPrompt[j] = commandline[i];
          j++;
      }
    stripCRLF(userPrompt);
    userPrompt[j]   = ' ';
    userPrompt[j+1] = '$';
    userPrompt[j+1] = '\0';
	
}


void execute(char commandline[])
{
    int i,j, errorCode=-1;
    char filename[200];
    char scriptcommand[500];
    FILE *ptr;

    // filename in the second word

    for(i=6,j=0;i<200 && commandline[i]!='\0';i++)
      if (commandline[i]!=' ') 
      {
          filename[j] = commandline[i];
          j++;
      }

    filename[j] = '\0';

    // Verify the file is good

    ptr = fopen(filename,"rt");
    if (ptr == NULL)
    {
        printf("Invalid script file\n");
        return;
    }

    // Process the script file to completion

    fgets(scriptcommand,500,ptr);
    stripCRLF(scriptcommand); // more commonly know as CHOMP
//    errorCode = interpreter(scriptcommand);

    while(!feof(ptr) && errorCode != -1)
    {
        fgets(scriptcommand,500,ptr);
        stripCRLF(scriptcommand);
//        errorCode = interpreter(scriptcommand);
    }

    fclose(ptr);
}

/*******************************************************************************/
/*	Function setQuanta: Sets the Quanta for the CPU's Run-time-environment
/*	
/*	Input: 
/*      - qanta(global variable)   1 quanta = 1 instruction per CPU cycle
/*      - commandline[]: user input
/*******************************************************************************/


int setQuanta(char commandline[])
{   // Reading ASCII char as an integer: [ASCII]49 = [int]1 and [ASCII]57 = [int]9
    int j;												// K decides where the character is placed

	quanta = 0;											// Reser to 0 for the algorithm to work
    for (j=6; j<30 && commandline[j]=='\0'; j++)		// Find the 1st non ' ' 

	stripCRLF(commandline); 
// We get kicked out of this loop when commandline[j] = NULL
	j++; 
	while(commandline[j])
		{
			quanta = (quanta*10);					// Shift old value << 1 dec. place
			quanta = quanta + (commandline[j]-48);    // Add LS digit
			j++;
		}

    printf("Quanta has been set to '%d'\n", quanta);
	return quanta;
}


/*******************************************************************************/
/*	Function Pushterminate: Pushes the PID of a PCB into the Terminate Queue
/*      Since we only allow 10 processes to run at the same time, 
/*      the terminate Q can allow 10 members, and it gets cleard every CPU cycles
/*      the list can never overflow
/*
/*	Inputs:
/*		- PID: PID of the process ready for termination 
/*******************************************************************************/
void PushTerminate(PID)
{
    int i;
	for(i=0; i<9;i++)					    // Now scan the Queue from start to end looking for next empty spot
	    {
    		if(terminateQ[i] == -1)         // If no PCB at Queue location 
                {
                    terminateQ[i] = PID;
                    break;
        	     }
        }
}



/*******************************************************************************/
/*	Function PushReadyQueue: Pushes a PCB into the Ready Queue
/*
/*	Inputs:
/*		- readyHead: Pointer to the Head of the Ready queue (declared globally)
/*
/*	Output:
/*		- Pointer to the progream in ready Queue (used as a PID) 
/*******************************************************************************/

int PushReadyQueue(PID)
{
	int i=0;
    if(readyQ[readyHead] == -1)             // Therefore readyHead is pointing to an empty spot
      {
        readyQ[readyHead++] = PID;          // Pushes the PID in the Q
        if(readyHead == 10)                 // Since we only allow up to 10 PCBs, condition implies that
           readyHead = 0;		            // we reached the end of the Queue; need to wrap arround to find space
// For debugging
/*for(i=0; i<10;i++)
	printf("ReadyQ[%d] = %d\n", i, readyQ[i]);

printf("ReadyH = %d, ReadyT = %d\n", readyHead, readyTail);
*/    	return 0;                           // Exit function and return 0 = ok!
      }

		return -1;						    // No empty spot in the list; returns error code
}// End PushReadyQueue





/*******************************************************************************/
/*	Function PopReadyQueue: Pops a PCB from the ready Q
/*
/*	Inputs:
/*		- readyTail: (global) pointer to the tail of the ready Queue
/*		
/*	Output:
/*		- PID of the poped element 
/*******************************************************************************/


int PopReadyQueue()
{
	int popedPID;
	popedPID = readyQ[readyTail];					    // Tail points to the last element of the Q

    if(popedPID != -1)
        {
            readyQ[readyTail++] = -1;                     // Empty spot in the Q
	      
        // Since we implemented a circular list and porgrams are queued on a FIFO basis form the 
        // lowers pointer (0) to the highest (9) then looping back 
	        if(readyTail == 10)
                readyTail = 0;              	       	// Means we reached the end of the Queue => 
        }                                               // need to wrap arround to see if there is space
	  
	return popedPID;	    						    // return Queue pointer as a PID
} // End PopReadyQueue()


/*******************************************************************************/
/*	Function PopLoadingQueue: Pops a PCB from the loading Q
/*	(might get merged with popReadyQ if i have time)
/*	Inputs:
/*		- loadingTail: (global) pointer to the tail of the ready Queue
/*		
/*	Output:
/*		- PID of the poped element 
/*******************************************************************************/


int PopLoadingQ()
{
	int popedPID;
	popedPID = loadingQ[loadingTail];					    // Tail points to the last element of the Q

    if(popedPID != -1)
        {
            loadingQ[loadingTail++] = -1;                     // Empty spot in the Q
        // Since we implemented a circular list and porgrams are queued on a FIFO basis form the 
        // lowers pointer (0) to the highest (9) then looping back 
	        if(loadingTail == 49)
                loadingTail = 0;              	       	// Means we reached the end of the Queue => 
        }                                               // need to wrap arround to see if there is space
	  
	return popedPID;	    						    // return Queue pointer as a PID
} // End PopReadyQueue()

/*******************************************************************************/
/*	Function createPCB: Creates a PCB for 'program'
/*
/*	Inputs:
/*		-RAMptr: "pointer" to start of program in RAM (since RAM is an array; address = int)
/*		-filename: Name of the program
/*		-lenght: number of RAM cells occupied by program (number of commands)
/*
/*	Output: 
/*		PID: not a real PID, since it represents where in the Ready Q the 
/*		PCB is sitting, but good ennough to refer to a process since
/*		only 1 process can be at one PCB at a time
/*
/*  Retunr: PID or -1 if error
/*******************************************************************************/


int createPCB(char filename[], int length)
{	
	int i,k;
	int PID = -1; 

 	// Since we have an array as Q, we need to scan for an 'empty' PCB
	for (i=9; i>=0; i--)
		{
			if(PCB[i].length == 0)
				PID = i;
		}                                  
	
    if(loadingQ[loadingHead] == -1)                             // If there is room in the readyQ
        {
        	for(i=0; i <=50; i++)                   	        // Insert File name; could have used strncpy.
    	     {PCB[PID].filename[i] = filename[i];}		
        	//PCB[PID].start = RAMptr;							// Address in RAM with start of the program	
        	//PCB[PID].current	= RAMptr;						// Since not executing yet, current = start 
        	PCB[PID].length = length;							// Given to us when we looked for empty space
	        memset(PCB[PID].memory, '\0', sizeof(PCB[PID].memory));		// Formats memory
			PCB[PID].current = 0;	 							// Instruction to be executed next
			for(k=0; k<=i; k++)
				{
					PCB[PID].pageTable[k].inRAM = 0; 			// None of the pages are in RAM
					PCB[PID].pageTable[k].wasModified = 0;  	// Cuz i'm sure we'll need it eventually 	
					PCB[PID].pageTable[k].baseAddress = -1; 	// To notify that it never got loaded in RAM
				}
            printf("'%s' has been assigned PID: '0x0%d'\n", filename, PID);
         }
    else
        {
            printf("Too many programs in the loading Q at the moment...\n");
            printf("Try to run this again later!\n\n");
        }

	return PID;
}// End createPCB





/*******************************************************************************/
/*	Function storeInRAM: 
/*		- Creates Backing store (/BACKINGSTORE/PROG.PCB.BS
/*		- Calls functions to initiate PCB
/*		- Push new PCB in Loading Queue
/*
/*	Inputs: 
/*	   - filename[]: contains the script file's name
/*
/*  Returns:
/*     - PID (redy Queue index) or -1 if not ennough room in RAM to store
/*******************************************************************************/
int storeInRAM(char filename[])
{
    int i,j,k;
    int scriptLenght=0;		
    char scriptcommand[100];
	char filenameBS[100];
    int PID;
    FILE *ptr;						// Script file
	FILE *BS;						// Created Backing Store
	struct PCBdef tempPCB;
	struct bs_rec temp_bs_rec;
  


    if((ptr = fopen(filename,"rt")) == NULL)            // Open the file and check if valid
        {
            printf("Can't open script file\n");         // Notyfy user and return error code -1
            return -1;                                  
        }

    while(!feof(ptr))	            					// Compute the space needed in RAM
    {
        fgets(scriptcommand,100,ptr);                   // Counts the rows filled with text of file
	if(scriptcommand[0] != '\n')scriptLenght++;		    // To avoid counting empty rows
    }
	scriptLenght--;
    printf("Program in '%s' is %d instructions long.\n",filename,scriptLenght);
	if(scriptLenght > 400)
		printf("WANNING!! Page table can only accomodate 400 lines of scritp. Your script got truncated at 400!!\n");

/**************************************************************************************/
/*THE BACKING STORE
/**************************************************************************************/
//    1. Create a PCB attached to a temporary pointer.
	printf("Creating PCB\n");

	PID = createPCB(filename, scriptLenght);			// Creates temporary PCB
	if(PID == -1)                                       // Warns user of cause of error
         return -1;                                     // If PID couldn't be created, Exit function

/**************************************************************************************
   2. fwrite the struct in write mode into a text file having the same name as the script
       file but with the extension .BS (for Backing Store). Normally this file would be
       saved in a hidden area called the backing store, but your backing store will be a
       subdirectory named BACKINGSTORE. You can put this subdirectory in the
       current directory.*/
/**************************************************************************************/

// Directory Backingstore got created in main()
	strcpy(filenameBS,"BACKINGSTORE/"); 
	k = 13;
    for(i=0;i<100 && filename[i]!='\0';i++)				// Loops through i, starting at the first space until we get to '\0'
        {
			filenameBS[i+k] = filename[i];
		}
// TO allow many programs to run simultaneously, the bs files will have the structure:
//   "PROGRAM.PID.BS"   Since every running program gets assigned a unique PID when running,
// Every entry should be unique.
		filenameBS[i+k]	  = '.';
		filenameBS[i+k+1] = (PID+48);
		filenameBS[i+k+2] = '.';
		filenameBS[i+k+3] = 'B';
		filenameBS[i+k+4] = 'S';
		filenameBS[i+k+5] = '\0';

// Look if there already exists an BS entry with same name. Since we are appending when writing, need to delete it
		if(!(remove(filenameBS)))
			{
				printf("WARNING!!\nFound an entry named '%s' in the BS\n",filenameBS);
				printf("Old entry got deleted and will be replaced by this one.\n\n");
			}

		if(!(BS = fopen(filenameBS, "w+")))				// Create the Backing store
			printf("Error creating Backing store\n"); 	
		else
			printf("Creating '%s' for PID0x0%d\n...\n", filenameBS, PID);

/***************************************************************************************
   3. Load the script line-by-line into the page_rec structure.*/
/**************************************************************************************/
	i = 0;												// I will count the pages
// The algorithm could allow unlimited pages, but out page table has a max of 100 entries, so we need to 
// cap the pages at 100

	rewind(ptr);							            // Rewinds ptr to start of the file
    while(!feof(ptr) && j < 100)
	    {												// Extract file commands and store in struct
			j = 0;
// First we extract the comand
			fgets(scriptcommand,100,ptr);
// Skip if its an empty row and read the next one
			while(scriptcommand[0] == '\n' && !feof(ptr)) // To avoid picking up empty spaces
				fgets(scriptcommand,100,ptr);
		   	stripCRLF(scriptcommand); 		        // Clean up command
// Copy into struct
			for(k=0;k<100;k++)temp_bs_rec.command1[k]=scriptcommand[k];			

// If we reached EOF, mark J and format command in struct
			if(feof(ptr)){j++;temp_bs_rec.command1[0] = '\0';}


// Repeat the operation 4 times in a row, then start over    
			fgets(scriptcommand,100,ptr);
			while(scriptcommand[0] == '\n' && !feof(ptr)) // To avoid picking up empty spaces
				fgets(scriptcommand,100,ptr);
		   	stripCRLF(scriptcommand); 		        // Clean up command
			for(k=0;k<100;k++)temp_bs_rec.command2[k]=scriptcommand[k];			
			if(feof(ptr)){j++;temp_bs_rec.command2[0] = '\0';}

			fgets(scriptcommand,100,ptr);
			while(scriptcommand[0] == '\n' && !feof(ptr)) // To avoid picking up empty spaces
				fgets(scriptcommand,100,ptr);
		   	stripCRLF(scriptcommand); 		        // Clean up command
			for(k=0;k<100;k++)temp_bs_rec.command3[k]=scriptcommand[k];			
			if(feof(ptr)){j++;temp_bs_rec.command3[0] = '\0';}

			fgets(scriptcommand,100,ptr);
			while(scriptcommand[0] == '\n' && !feof(ptr)) // To avoid picking up empty spaces
				fgets(scriptcommand,100,ptr);
		   	stripCRLF(scriptcommand); 		        // Clean up command
			for(k=0;k<100;k++)temp_bs_rec.command4[k]=scriptcommand[k];			
			if(feof(ptr)){j++;temp_bs_rec.command4[0] = '\0';}
									
/***************************************************************************************
   4. Repeat steps 2 & 3 for all the lines in the script. If the last page did not use all 4
       arrays in the struct, make sure to initialize the unused arrays to '\0'*/
/**************************************************************************************/

			// If we haven't hit 4 EOFs in a row, write the page on the BS
			if(j!=4)
				{
					fwrite(&temp_bs_rec, sizeof temp_bs_rec, 1, BS);
					i++;// Increment page counter
				}

	    }// End while(!eof)
		printf("Total of %d pages created and stored.\n",i);

/**************************************************************************************
   5. When done, create a field in the PCB called FILE *BACKINGSTORE that will
       point to that .BS file. Created another field in the PCB called long int
       MAXPAGES. Save the the total number of created pages for that .BS file.
/**************************************************************************************/
		rewind(BS);
		PCB[PID].BACKINGSTORE = BS;
		PCB[PID].MAXPAGES = i;   // Counter trigered at every page creation

		fclose(ptr);


// Warn user if program is = or longer than Page table size
		if(j==100)printf("Reached maximum of 100 instructions. Anything beyond has been ignored!\n\n");


/**************************************************************************************
   6. Insert the PCB into the Loading queue. (Loading Q implemented using an array[9]
/**************************************************************************************/
	if(loadingQ[loadingHead] == -1)         // Empty spot are marked by -1
      {
        loadingQ[loadingHead] = PID;        // Pushes the PID in the Q
        if(loadingHead == 49)                  // Since we only allow up to 49 PCBs, condition implies that
                loadingHead = 0;              // we reached the end of the Queue; need to wrap arround to find space
        else 
	        loadingHead += 1;                     // Pointer to head of ready Q is incremented s.t. it points to first empty spot
	  }// End if there was room in the Q
	else
		printf("WARNING!\n Too many programs in the loading Q at the moment. Try again later!\n");
	
   	printf("Process initiation completed!\n\n");                                // Implemented while debugging, kept Just for fun...

// PCB is in Loading Queue and will be executed during CPU run-time; 
	return PID;

}// END of storeInRAM function












/*******************************************************************************/
/*	Function run: 
/*      - Initiates the command 'RUN'
/*      - Extracts filename(s) from user input
/*      - calls soreinRAM (described above) for every filenames found
/*	
/*	Inputs: 
/*	   - userInput[]: contains the script file's name
/*******************************************************************************/

int run(char userInput[])
{
	int i, j=0;
	char filename[200];
	int canRun = 0;

// Count if there is still room to run more processes

	for(i=0; i<10; i++)
		if(PCB[i].length == 0)
			{canRun++;
//		printf("canRun = %d\n",canRun);
}

// Extract the file name	    
    for(i=4;i<200 && userInput[i]!='\0';i++)	// Loops through i, starting at the first space until we get to '\0'
        {
            filename[j] = userInput[i]; 	    // Extract the filename from the user input
            j++;
            if(userInput[i]==' ')		        // Means there is more than one file name
		        {                               // Need to store the first one in RAM
		            filename[j-1] = '\0';			// Complete the file name
					if(canRun-- > 0)
	          	        storeInRAM(filename);		// Extract script script data, store in RAM and create PCB
					else
						printf("Reached MAX processes the OS can handle\n RUN request for '%s' has been ignored!\n\n", filename);      
			
		            j = 0;				        // Resets j counter for next filename
		        }// End if userinput = ' ' 
	    }// End for loop

// For the last file name of the RUN command; j counter hasen't been reset = file script not processed yet
	if(j!=0)  
	{
	   filename[j] = '\0';			            // Complete the file name
		if(canRun>0)
	       storeInRAM(filename);	    	        // Extract script script data, store in RAM and create PCB
		else
			printf("Reached MAX processes the OS can handle\n RUN request for '%s' has been ignored!\n\n", filename);      
	}    

	return 0;
}// End run();

/*******************************************************************************/
/*	Function: memManager (Memory Manager)
/*	
/*	Input: PID of program (PID.current will be used)
/*	FRAMESIZE: Global variable, can be updates only at launch
/*
/*	What it does:
/*      1. Loads PID.current page  into RAM
/*			if new PCB, loads first page
/*			if called surring run-time, loads next page to be executed
/*      2.A.  Page is loaded using first fit strategy. 
/*		2.B   If all of RAM is full then the manager removes a page from RAM from
/*		the PCB at the tail of the ready list. If that PCB has NO pages in RAM we 
/*		go to the second to last PCB, and so on until we find a PCB with a page in RAM. 
/*			Overwrite the oldest cell, which should be cell 0. If that page is already
/*      not in RAM, we will then overwrite the next oldest which is cell 1, and so on
/*      until we fine one. We only overwrite one page since we only need one page.
/* 			Then updated the victim's page table and overwrite the frame with the new commands.  
/*      3. PCB is added to the tail of the Ready queue.
/*			if new PCB, got removed from loading Q
/*			if running PBC, looses quanta and goes to the back of the readyQ
/*******************************************************************************/

void memManager(int newPID)
{
	int victimePID;
	int RAMptr = -1;
	int current;
	int i,j,k;
	int tempReadyTail = readyTail;


	printf("Starting Memory Manager for PID: 0x0%d\n",newPID);
// Look for an empty FRAME  in RAM

    for(j=0; j<=(100-FRAMESIZE); j++)				    // Scan RAM memory array for empty spot
     {				
		if(FREERAM[j] == '\0')				            // If start of frame = \0, whose frame is empty
		    {	                                           
		        RAMptr = j;	            				// Save location of start of empty FRAME
			    printf("Found an empty FRAME at address RAM0x");
	            if(j < 16)printf("0");                  // Created for debugging, kept just for fun :)
				printf("%x\n",RAMptr);
		        break;                              // Breaks out of for loop s.t. scanningRAM=0 can take effect
	        }// End if found a spot in RAM	
		}//End forJ loop


// If we didn't find a spot in RAM; we need to loook for a victim.
// We push the PCB into the ready Q now, s.t. if it's the only one
// running and hogging the whole RAM array, it will become its own victim.


// First we figure out how many pages we need to extract
	int PPF;										// # Pages per frame 
	PPF = FRAMESIZE/4;								// Since 4 instructions per page (default 1 PPF if FRAMESIZE =4)
	struct bs_rec data;								// In order to lead a frame, we need to extract PPF pages
	 	
// Then look for a victime

							// So we are alone


	for(i=0; (i<10 && (RAMptr == -1)); i++)				// Loop through until we find a spot 			
	{
		victimePID = readyQ[tempReadyTail++];			// Pop the last guy in the readyTail with a temp pointer
		if(tempReadyTail == 10)tempReadyTail = 0;		// Since circular Q

		if(victimePID != -1)
		printf("RAM is full at the moment... Attempting to steal space from PID0x0%d\n",victimePID);

		if(victimePID == -1)
			{
				victimePID = newPID;							// Flag ourselves as a Victime
				printf("RAM is full at the moment... Removing some of PID0x0%d's pages\n",victimePID);
			}

// Now that we found a victime, or ourselves
		printf("Booting off PID:0x0%d's\n",victimePID);
		for(k=0; (k<100 && (RAMptr == -1));k++)			// Scan the victim's page table	
			{
				if(PCB[victimePID].pageTable[k].inRAM)	// Oldest page table in RAM
					{
						RAMptr = PCB[victimePID].pageTable[k].baseAddress;	// Steal his spot in RAM
						for(j=0;j<PPF;j++)							// We knock out one page per pass
						{
							printf("Page #%d out of RAM0x",k);
							if(RAMptr<16)printf("0");
							printf("%x\n",PCB[victimePID].pageTable[k].baseAddress);
							PCB[victimePID].pageTable[k].inRAM = 0;				// Then boot him out!!!
							PCB[victimePID].pageTable[k].baseAddress =0;
							k++;
						}
					}// End if
			}// End for k loop
	}
	printf("\n");
						



// Now RAMptr points to an empty frame         


// Extract the next page(s) to be loaded into RAM the next page


// We'll use 'current' to keep track of which page we are at and update the page table
	current = PCB[newPID].current/4;

// Then we extract the pages and store them remporarly into a data structure
		FILE *fptr = PCB[newPID].BACKINGSTORE;		// The file handle has been left open and rewinded :)

		long offset = current * 400;					// Calculate the offset
		fseek(fptr,offset, 0); 


// We read one page per pass until we reach the last page
		printf("Loading PID:0x0%d's\n",newPID);
		for(k=0; (k<PPF && current < PCB[newPID].MAXPAGES);k++)							
			{
				if(feof(fptr) )    // check for EOF
				{
					printf("Reached the end of the BS for PID0x0%d\n",newPID);
			    	break;
				}
				if(!(fread(&data, sizeof data, 1, fptr)))
					printf("Error reading from Backing Store.\n");
// Load page(s) in RAM starting at address RAMptr and update page table

			printf("Page #%d into RAM0x", current);			
			if(RAMptr<16)printf("0");
			printf("%x\n", RAMptr);
			   PCB[newPID].pageTable[current].inRAM = 1;
			   PCB[newPID].pageTable[current].baseAddress = RAMptr; 
               FREERAM[RAMptr] = strdup(data.command1);
				RAMptr++;
               FREERAM[RAMptr] = strdup(data.command2); 
				RAMptr++;
               FREERAM[RAMptr] = strdup(data.command3);
				RAMptr++;
               FREERAM[RAMptr] = strdup(data.command4); 
     			current++;
				RAMptr++;
// FOR DEBUGGING  
				// printf("k=%d\n, Command1: %s,Command2: %s,\nCommand3: %s, Command4: %s \n",k, data.command1,data.command2,data.command3,data.command4);
			}
			printf("\n");

//			rewind(fptr);


// FOR DEBUGGING  
//for(k=0;k<100;k++)
//	printf("RAM[%d] = %s\n", k, FREERAM[k]);

// The process is pushed into the ready Q externally
	    
}








/*******************************************************************************/
/*	Function: shell()
/*	
/*	The main command-line control loop
/*******************************************************************************/


int shell()
{
    int i, errorResult=0;
    char response;
    char userInput[200];
    int Running = -1;               // Tells the interpreter that the shell is sending the command 

    // Display the command-line prompt
    printf("%s", userPrompt);
    fgets(userInput,200,stdin);
	stripCRLF(userInput);
    errorResult = interpreter(userInput, Running);

    return errorResult;
}



/****************************************************/
/*	Function: switches()
/*
/*	inputs:
/*		argc 	= argv[] element count
/*		argv[] 	= contains all the arguments inputed
/*			argv[0]	= 	launchcpde
/*			argv[1] = 	parameter1
/*			argv[n] = 	parametern
/*			argv[n+1] =	null
/*
/*	outputs:
/*		Returns 0 if no inputs => Programm launches
/*		Returns 1 if an argument was inputed => Program terminates
/*		Returns -1 if illegal argument input => Program terminates
/*
/*	Function:
/*		Acts on decoded argument
/*
/*
/****************************************************/
int switches(int argc, char **argv)
{
	if(argc == 0)return 0;						//No launch option, function returns 0 to main
    char framesize[3];
	int i, j, k;							

	for( i = 1; i <= argc; i++ )				//Scan through arguments; argv[argc]=null
     {
       switch((int)argv[i][0] )					//Check for '-'
       {
       case '-':					    		//Found '-'; interprets the option
			switch((int)argv[i][1])				//scans the 1st letter of arguments after '-'
                {
                   case 'h':					//Found the option -h; displaying help
					   puts("\nOption '-h' => Displaying Launch Parameters\n"
					        "To launch the OS, input 'os' -[argument] from the command line.\n"
							"Type 'os -h' to display help file\n"
							"The operating system was NOT launched...\n\n");
							return 1;       	//Tells main that valid option was found
                            break;				//Not nescessary.
                    case 'f':					// Possible framesize
                        if (strncmp(argv[i], "-framesize",10) == 0)
                            { 
                            k = 0;												// K decides where the character is placed
							FRAMESIZE = 0;										// Reser to 0 for the algorithm to work
                            for (j=10; j<30 && argv[i][j]=='\0'; j++)			// Find the 1st non ' ' 
                            {}  
							while(argv[i][j])
								{
									if(k)										// Means that this is not the 1st time evaluating
										FRAMESIZE = (FRAMESIZE*10);				// Shift old value << 1 dec. place
			
									FRAMESIZE = FRAMESIZE + (argv[i][j]-48);    // Add LS digit
									k++;										// Notify that we tampered with FRAMESIZE in case of next pass

									j++;   
								}
							concurrent = 100/FRAMESIZE;
							if(FRAMESIZE > 100)
								{
									printf("\nWARNING!!! Only 100 slots in RAM!!\n");
									printf("To avoid instabilities, framesize cannot exceed 100!\n"); 
									FRAMESIZE = 100;
								} 

							if(FRAMESIZE%4)
								{
									printf("\nWARNING!!! FRAMESIZE must be a multiple of 4!\n");
									printf("Regressing to an adequate value\n"); 
									FRAMESIZE -= FRAMESIZE%4;

								}                          
							printf("FRAMESIZE has been set to: %d\n\n", FRAMESIZE);

								
                            }// End switch
                            return 0;
                            break;
                   default:						//If no valid launch parameter have been found
				 		printf( "-%c is not a valid launch parameter\n", argv[i][1]);
                        return -1;				//Informs main that user input illegal launch parameter
                        break;
                 }// End of scanning -[parameter]

			break;// End of case '-'; no legal parameter calls have been found

       default:		    						//If parameter doesn't start with '-'; illegal
	   		puts("Illegal syntax: launch parameter must start with '-'\n"
				 "OS was not launched...\n\n");
			return -1;
            break;
		}// End of switch loop scanning parameters
      }// End of for loop scanning parameters
return -1;				    				    //The program shouldn't be able to get here
}//End of function switches()







/*******************************************************************************/
/*	Function: TaskSwitchIn(Running, CPUptr)
/*
/*
/*	Inputs:
/*		- Running: 'pointer' (int) to the PBC that got poped from ready Q and gets to run
/*		- CPUptr: real pointer to the CPU => defined as a strucu in main()
/*
/*	Outputs:
/*		None
/*
/*	Function: Transfer the information from PCB to CPU
/*	Returns: None
/*
/*	*remember Datatypes definitions:
/*     PCB(char filename[50], int start, int length, int current, char *memory[20], struct PCB *next)
/*	CPU(int quanta, int returnCode,	int pc,	char *memory[20])
/*******************************************************************************/

void TaskSwitchIn(struct cpu *CPUptr, int Running)
{
	int i;	
	CPUptr->quanta = quanta;					// Reset quanta before starting job
	CPUptr->returnCode = 0;						// Reset CPU return code
	CPUptr->pc = PCB[Running].current;      	// Set the PC to where it left off
}


/*******************************************************************************/
/*	Function: TaskSwitchOUT(Running, CPUptr)
/*
/*
/*	Inputs:
/*		- Running: 'pointer' (int) to the PBC that got poped from ready Q and gets to run
/*		- CPUptr: pointer to the CPU => defined as a strucu in main()
/*
/*	Outputs:
/*		None
/*
/*	Function: Transfer the information from CPU to PBC
/*	Returns: None
/*
/*	*remember Datatypes definitions:
/*     PCB(char filename[50], int start, int length, int current, char *memory[20], struct PCB *next)
/*	CPU(int quanta, int returnCode,	int pc,	char *memory[20])
/*******************************************************************************/

void TaskSwitchOut(struct cpu *CPUptr, int Running)
{
    PCB[Running].current = CPUptr->pc;                 	// Records PC
}





/****************************************************/
/*	Function: formatRAM()
/*
/*	inputs:
/*		char *FREERAM[100] (A global variable
/*
/*	outputs:
/*		NONE
/*
/*	Function:
/*		Initializes FREERAM array by formating every
/*		cells (sets every elements to NULL)
/*
/*	Returns:
/*		0 if everything went well
/*		1 if an error occured
/*
/****************************************************/
int formatRAM(void)
{
	puts("Formating RAM...");
	memset(FREERAM, '\0', sizeof(FREERAM));
	puts(" ...Done!\n\n");
	return 0;
}// End of function formatRAM()




/*******************************************************************************/
/*	Function: login
/*
/*	Inputs: 
/*      - pass.txt: External password file conatining 
/*          NAME1
/*          PASSWORD1
/*          NAME1
/*          PASSWORD2
/*          [...]
/*          END
/*
/*	Function:
/*		Prompts for a name and password whitch are matched with an encrypted text file
/*		containing all valid credentials
/*
/*	Returns:
/*		0 Invalid Name or Password
/*		1 Valid login credentials
/*
/*******************************************************************************/
int login(void)
{
	FILE *fp;
	int i = 0;	
	int havename;
	int havepassword;
	
    struct credentials
	{
		char name[15];
		char pass[15];
	}user, fromfile;
	memset(user.name, '\0', sizeof(user.name));		// Formated to avoid collision login with existing memory values
	memset(user.pass, '\0', sizeof(user.pass));		// Formated to avoid collision login with existing memory values

	logo();									        // Displays OS logo... functionally useless

	printf("\nLogin name: ");						// Name is restricted to 15 characters to avoid iostream buffer overflow
	fgets(user.name, 15, stdin);                    // Get user input name

	printf("Password: ");                           // Get user input password
	fgets(user.pass, 15, stdin);					// Password also restricted to 15 characters

	if((fp = fopen("pass.txt","r"))==NULL)		    //Open the password file
		{
			puts("Cannot open password file!\n"		
				 "Please make sure you have a file named 'pass.txt'\n"
				 "containing login credentials in this directory.\n");
			return 0;
		}

/********************/
/*	Authentication	
/********************/


 	while (feof(fp) == 0)
  	{
		memset(fromfile.name, '\0', sizeof(fromfile.name));		// Format filename buffer avoid accidental
                                                                // login with existing memory values
    	fgets(fromfile.name, 15, fp);		        			// Read next record on pass.txt file
		havename = 1;	                						// Assume a valid user name until proven wrong
		for(i = 0; i < sizeof(user.name); i++)	        		// Check if valid user name
    		{
    			if(fromfile.name[i] != user.name[i])	        // Character by character analysis
    				{
                        havename=0;			                    // Any divergence between the 2 values 
						break;                                  // kicks the porogram out of this loop
                    }// End of looking for invalid name
		    }// End for(scanning name)

		if(havename)		                					// Still have valid name after matching with file values
			{							                        // We have to look for valid password
			memset(fromfile.pass, '\0', sizeof(fromfile.pass));	// Start by formating the password buffer
			fgets(fromfile.pass, 15, fp);	        			// Extract password string associated with name
			havepassword = 1;
			if(fclose(fp))puts("File close error.\n");	    	// Since user names unique, pass file is no longer needed

			while(havepassword)
				{
					for(i = 0; i < sizeof(user.pass); i++)	    // Scan for divergence between record and input passwords
						{
							if(fromfile.pass[i] != user.pass[i])// If != detected, invalid credentials = return 0
								{
									puts("\nInvalid password!\n\n");
									return 0;	                // Exit the function
								}
						}				
					puts("Loging in... ...\n");                 // Otherwise
					return 1;				                    // Code can only get here if divergence not detected
				}// End of havepassword
			}// End of havename; wrong password; function should have already terminated
	}// Reached EOF without finding a valid user name

	puts("\nInvalid user name\n\n");
	if(fclose(fp))puts("File close error.\n");	    			//Close the password file
	return 0;								                    //Return 0 by default
}//End of login script



/*******************************************************************************/
/*	Function: Command line interpreter
/*******************************************************************************/

int interpreter(char userInput[], int Running)
{
    char response;
    int i, j;
    char extraChars[10];
	int ILLEGAL = 0;
	int ErrorReturn;

    if (verbose)
    {
        printf("Verbose mode: Confirm execution of %s. (y/n): ",userInput);
        scanf("%c",&response);
        fgets(extraChars,10,stdin);
	stripCRLF(extraChars);
        if (response == 'N' || response == 'n')
        {
            printf("Operation cancelled\n");
            return 0;
        }
    }


/***************************************************/
// Assignment 4 additions
/***************************************************/
// Running is passed from the entity who issued the instruction. If it came from the CPU, 
// running is the PCB of the running process (already implemented for keeping 
// track of memory variables)
// If it came from the Shell, running = -1 (illegal PCB value) so these instructions are ignored

	if (strncmp(userInput,"OPEN",4)==0)
	{
		if(Running == -1)
			{
				printf("WARNING, This instruction can NOT be launched from the Shell!\n");
				ILLEGAL = 1;							// Flags the system to avoid this instruction being sent out to the real OS shell
			}
		else
			{
			ErrorReturn = GetRessourceIndex(userInput, Running);
			return ErrorReturn;			
			}	
	}

	if (strncmp(userInput,"WRITE",5)==0)
	{
		if(Running == -1)
			{
				printf("WARNING, This instruction can NOT be launched from the Shell!\n");
				ILLEGAL = 1;							// Flags the system to avoid this instruction being sent out to the real OS shell
			}
		else
			{
			ErrorReturn = WriteToFile(userInput, Running);
			return ErrorReturn;			
			}	
	}
	
	if (strncmp(userInput,"CLOSE",5)==0)
	{
		if(Running == -1)
			{
				printf("WARNING, This instruction can NOT be launched from the Shell!\n");
				ILLEGAL = 1;							// Flags the system to avoid this instruction being sent out to the real OS shell
			}
		else
			{
			ErrorReturn = closeINDEX(userInput, Running);
			return ErrorReturn;			
			}	
	}


	if (strncmp(userInput,"DELETE",6)==0)
	{
		if(Running == -1)
			{
				printf("WARNING, This instruction can NOT be launched from the Shell!\n");
				ILLEGAL = 1;							// Flags the system to avoid this instruction being sent out to the real OS shell
			}
		else
			{
			ErrorReturn = deleteFile(userInput, Running);
			return ErrorReturn;			
			}	
	}




    else if (strcmp(userInput,"EXIT")==0 || strcmp(userInput,"LOGOUT")==0)
        return -1;
    else if (strcmp(userInput,"CLS")==0)
    {
        for (i=0;i<30;i++) printf("\n");
    }
    else if (strcmp(userInput,"HELP")==0)
    {
        printf("\n\nmyshell Command Set\n\n");
        printf("EXIT            to terminate the shell\n");
        printf("CLS             to clear the screen\n");
        printf("HELP            displays this help screen\n");
        printf("VERBOSE ON/OFF  to toggle verbose mode\n");
        printf("SET VAR VALUE   add a variable to shell memory\n");
        printf("GET VAR         see the value of a shell memory var\n");
        printf("SCRIPT FILENAME execute a script\n");
	printf("QUANTA [int]	sets quanta time to [int] (default = 3)\n");
        printf("PROMPT VALUE    change your prompt\n");
        printf("\n");
    }
    else if (strcmp(userInput,"VERBOSE ON")==0)
        verbose = TRUE;
    else if (strcmp(userInput,"VERBOSE OFF")==0)
        verbose = FALSE;
    else if (strcmp(userInput,"GET")==0)
    {
        printf("Shell Memory Variables\n");
        for(i=0;i<100 && shellMemoryVar[i]!=NULL;i++)
        {
            printf("%s=%s\n",shellMemoryVar[i],shellMemoryData[i]);
        }
    }
    else if (strncmp(userInput,"GET",3)==0)
        displayMemVar(userInput);
    else if (strncmp(userInput,"SET",3)==0)
        setMemVar(userInput, Running);
    else if (strncmp(userInput,"PROMPT",6)==0)
        changePrompt(userInput);
    else if (strncmp(userInput,"SCRIPT",6)==0)
        execute(userInput);
    else if (strncmp(userInput,"RUN",3)==0)
	    run(userInput);
    else if (strncmp(userInput,"ECHO",4)==0)
	{
		for(i=5;i<200 && userInput[i]!='\0';i++)
			{putchar(userInput[i]);}
		printf("\n");
	}
   else if (strncmp(userInput,"QUANTA",6)==0)
	setQuanta(userInput);
   else
       if(!ILLEGAL)system(userInput);

    return 0;
}



/*******************************************************************************/
/*	Function:  createFS()
/*******************************************************************************/

int createFS()
{

	int i;

// First create a directory for the disk
	mkdir("FILESYSTEM", 0777);// File system directory mounting point

// Delete previous File system if it exists (NOT REQUIRED, just to let the user know what is hapenning)
	if(!(remove("FILESYSTEM/OSFAT.txt")))printf("Deleting old FAT table (Found FILESYSTEM/OSFAT.txt)\n");
	if(!(remove("FILESYSTEM/OSFILES.txt")))printf("Deleting old HDD content (Found FILESYSTEM/OSFILES.txt)\n");


// Create new file system. 
// *FAT and *FILE are real file pointers used by the OS to manage the virtual environment
// Open for reading and writing and create the file if it does not exist. 
		if(!(FAT = fopen("FILESYSTEM/OSFAT.txt", "w+")))
			{
				printf("Error creating FAT file. \n");
				return -1; 	
			}
		else
			printf("Creating FAT table... DONE!\n");



		if(!(FILES = fopen("FILESYSTEM/OSFILES.txt", "w+")))
			{
				printf("Error creating OS Virtual HDD. \n");
				return -1; 	
			}
		else
			printf("Creating OS Virtual HDD... DONE!\n");


// Finally format virtual HDD by blocks of 10 bites at a time

	printf("Formatting virtual HDD...");
		
	for(i = 0; i<100; i++)
		fprintf(FILES, "0000000000");

	printf("DONE! (1 Kb formatted in total)\n");

	return 1; // Everything went well
}

/*******************************************************************************/
/*	Function:  PushWaitingQ
/*******************************************************************************/

int PushWaitingQ(int PID, int INDEX)
{


    if(waitingQ[INDEX][waitingHead[INDEX]] == -1)             // Therefore readyHead is pointing to an empty spot
      {
        waitingQ[INDEX][waitingHead[INDEX]++] = PID;          // Pushes the PID in the Q
        if(waitingHead[INDEX] == 49)                 // Since we only allow up to 10 PCBs, condition implies that
           waitingHead[INDEX] = 0;		            // we reached the end of the Queue; need to wrap arround to find space

/*For debugging
int i;
for(INDEX = 0; INDEX<5;INDEX++)
{
	for(i=0; i<20;i++)
		printf("waitingQ[%d][%d] = %d\n",INDEX, i, waitingQ[INDEX][i]);
	printf("waiting[%d]H = %d, waiting[%d]T = %d\n",i, waitingHead[INDEX],i, waitingTail[INDEX]);
} 
*/



   	return 1;                           // Exit function and return 0 = ok!
	}
	return -1;						    // No empty spot in the list; returns error code
}// End PushReadyQueue



int PopWaitingQ(int INDEX)
{
	int popedPID;
	popedPID = waitingQ[INDEX][waitingTail[INDEX]];					    // Tail points to the last element of the Q

    if(popedPID != -1)
        {
            waitingQ[INDEX][waitingTail[INDEX]++] = -1;                     // Empty spot in the Q
        // Since we implemented a circular list and porgrams are queued on a FIFO basis form the 
        // lowers pointer (0) to the highest (49) then looping back 
	        if(waitingTail[INDEX] == 49)
                waitingTail[INDEX] = 0;              	       	// Means we reached the end of the Queue => 
        }                                               // need to wrap arround to see if there is space
	  
	return popedPID;	    						    // return Queue pointer as a PID
} // End PopReadyQueue()






/*******************************************************************************/
/*	Function terminateprocesses: Terminates all processes in terminateQ
/*******************************************************************************/

// Close BS, delete BS
void TerminateProcesses()
{
	int i,j, k, t,PID;
	char filenameBS[100];

                               
    for(t=0; t <= 9; t++)                                 	// Look at every elements in the Q 
    {
    	PID = terminateQ[t];                                // Extract element to be terminates                           
        if(PID != -1)                                       // If we found a valid PID that needs termination 
        {
       		printf("Terminating PID: 0x0%d\n", PID);    	// Implemented while debugging, kept Just for fun...

	   // close file handles if there are any open 
		for (i=0;i<5;i++)
		{
			if(FileHandles[i][1] == PID)
			{
				printf("OS is releasing (for you) file handle #%d\n",i);
				FreeResourceIndex(i);
			}
		}
// Delete shell variables used by process
            for(k=0; (k <=20 && PCB[PID].memory[k] != NULL); k++)           // Scan every non empty memory cells of the PCB
            {              
            	for(j=0;(j<100 && shellMemoryVar[j] != NULL);j++)           // Comapre every SHELL memory variable with the PCB's variables
                {
                	if (strcmp(PCB[PID].memory[k],shellMemoryVar[j])==0)    // When we find a match
                    {
                    	printf("DELETING '%s=%s' from Shell Memory\n",shellMemoryVar[j],shellMemoryData[j]);
                    	free(shellMemoryVar[j]);            // Free the memory
                        shellMemoryVar[j] = '\0';           // And set to NULL
                        free(shellMemoryData[j]);
                        shellMemoryData[j] = '\0';
                        break;
                    }
          		}
          	}

// Delete PCB's record of used variables
   	        memset(PCB[PID].memory, '\0', sizeof(PCB[PID].memory));		// Then format PCB memory

// Free RAM used by process
			for(j=0; j<100; j++)
				if(PCB[PID].pageTable[j].inRAM)
				{
					FREERAM[PCB[PID].pageTable[j].baseAddress] = NULL;     	                    
					PCB[PID].pageTable[j].baseAddress = -1;
					PCB[PID].pageTable[j].inRAM = 0;
				}

// Then delete the backing store file
			strcpy(filenameBS,"BACKINGSTORE/"); 
			k = 13;
		    for(i=0;i<100 && PCB[PID].filename[i]!='\0';i++)				// Loops through i, starting at the first space until we get to '\0'
				filenameBS[i+k] = PCB[PID].filename[i];
			filenameBS[i+k]	  = '.';
			filenameBS[i+k+1] = (PID+48);
			filenameBS[i+k+2] = '.';
			filenameBS[i+k+3] = 'B';
			filenameBS[i+k+4] = 'S';
			filenameBS[i+k+5] = '\0';

			if(fclose(PCB[PID].BACKINGSTORE))
				printf("Error closing backingstore file %s", filenameBS);							

			if(remove(filenameBS))
				printf("Error Deleting backingstore file %s", filenameBS);							

// Then delete the rest of the data from PCB
			for(j=0; j <=20; j++)          
				PCB[PID].filename[j] = '\0';	        		
	        PCB[PID].start = 0;					      
            PCB[PID].current = 0;			 
            PCB[PID].length = 0;
			
          	printf("Done terminating PID: 0x0%d\n", PID);               // Implemented while debugging, kept Just for fun...
			terminateQ[t] = -1;
         }// End if PID = -1
                                           
    }
}

/*******************************************************************************/
/*	Function makeBuffer(int HDDpointer)
/*******************************************************************************/

void makeBuffer(int HDDpointer)
{
	int i;
	char byte[2];

	for(i=0; i<BUFFER && (i+HDDpointer) < 1000;i++)
	{
		fseek(FILES,(i+HDDpointer), 0);
		fgets(byte,2,FILES);		  		//Grab a byte
		if(byte[0] == '0')						// If the byte is empty
			{
				fseek(FILES,(i+HDDpointer), 0);
				fprintf(FILES, "9");				// copy 9 to indicate that the space is reserved for the file handle buffer 
			}
	// Otherwise, break in case we reached the next file. 
		else
			break;
	}// End for
}


/*******************************************************************************/
/*	Function GetRessourceIndex(char userInput[], int PIDRunning)
/*******************************************************************************/


// PID running is passed on from the CPU. If the command came from the SHELL, the interpreter already filtered it out 
// CALL SYNTAX: OPEN FILENAME AT INDEX
int GetRessourceIndex(char userInput[], int PIDRunning)
{
	int i,j;
	char byte[2];
	int INDEX;
	int HDDpointer;
	int countFATRecords;
	FILE *TEMPFAT;
	struct OSFAT existingFATentry, newFATentry;
	int fileExists = 0;

// for debugging
//for(INDEX=0;INDEX<5;INDEX++)
//	printf("FileHandles[%d][0] = %d\tFileHandles[%d][1] = %d\n”",INDEX,FileHandles[INDEX][0],INDEX,FileHandles[INDEX][1]);
	
	printf("EXECUTING '%s'\n",userInput);

// Extract the file name	 
    // skip white spaces after OPEN
    for(i=4;i<200 && userInput[i]!='\0' && userInput[i]==' ';i++);

    // copy filename
    for (j=0;j<20 && userInput[i]!='\0' && userInput[i]!=' ';i++,j++)
        newFATentry.filename[j]=userInput[i];   //printf("i=%d, userinput[%d]=%c\n",i,i,userInput[i]);}

	newFATentry.filename[j] = '\0';

	// Display error if filename too long
	if(userInput[i]!=' ')
	{
		printf("\nInstruction '%s' ignored!\n", userInput);
		printf("(Filename restricted to 20 characters!)\n\n");
		return -5;
	}

    // skip white spaces after the filename
	i++;
    for(;i<200 && userInput[i]!='\0' && userInput[i]==' ';i++);

	// Look for AT
	if(userInput[i++] != 'A' || userInput[i++] != 'T')
	{
		printf("\nInstruction '%s' ignored!\n", userInput);
		printf("(Illegal Syntax => should look like 'OPEN FILENAME AT INDEX'\n\n");
		return -5;
	}


    // skip white spaces after AT
	i++;
    for(;i<200 && userInput[i]!='\0' && userInput[i]==' ';i++);


	// Extract index #
	// Reading ASCII char as an integer: [ASCII]49 = [int]1 and [ASCII]57 = [int]9
	INDEX = 0;
// We get kicked out of this loop when commandline[j] = NULL
	while(userInput[i])
		{
			INDEX = (INDEX*10);						// Shift old value << 1 dec. place
			INDEX = INDEX + (userInput[i]-48);    	// Add LS digit
			i++;
		}
	if(INDEX < 0 ||INDEX > 4)
	{
		printf("\nInstruction '%s' ignored!\n", userInput);
		printf("(Illegal INDEX pointer => must be between [0..4])\n\n");
		return -5;
	}


// At this point, both the filename and the desired index numbers have been extracted


// Look if the requested file handle is available
	if(FileHandles[INDEX][0] != -1)
	{
		if(FileHandles[INDEX][1] != PIDRunning)
		{
			printf("Sorry, file '%s' is already opened by PID0x0%d! Try again later. \nOperation cancelled.\n...\n",existingFATentry.filename,FileHandles[INDEX][1]);
			INDEXwait = INDEX;
			return -2;
		}
		else
		{
			printf("Your process (PID0x0%d) already owns Filehandle#%d.\nTo avoid deadlock, operation cancelled!!\n\n",PIDRunning,INDEX);
			return -6;
		}			
	}

// If the code gets here, we know that the index is free.


	// Look up the filename and HDD location from the FAT too see if it exists


	rewind(FAT);
	// fileExists initialized at 0
	printf("\nLooking up FAT records for '%s'...\n",newFATentry.filename);
	// Scan FAT table until EOF or until found existing file name
 	while (feof(FAT) == 0 && fileExists == 0)
  	{
		memset(existingFATentry.filename, '\0', sizeof(existingFATentry.filename));		// Format buffer avoid accidental
		TEMPFAT = FAT;
    	if(!(fread(&existingFATentry, sizeof existingFATentry, 1, FAT)))
			{
				printf("Reached end for FAT\n");	// Read next record on OSFAT.txt file
				break;
			}
		else
			printf("-rwxrw-rw-\tPID0x0%d\t  %d\t%s\n",PIDRunning,existingFATentry.address, existingFATentry.filename);
	
		fileExists = 1;	                						// Assume File exists until proven wrong
		for(i = 0; existingFATentry.filename[i] != '\0'; i++)	// Check if File exists
    		{
    			if(existingFATentry.filename[i] != newFATentry.filename[i])	        // Character by character analysis
    				{
                        fileExists=0;			                // Any divergence between the 2 values
						break;                                  // kicks the porogram out of this loop
                    }// End of looking filename
		    }// End for(scanning filename)
	}// Reached EOF without finding a matching filename name
	printf("\n");

// Here, we have a free file handle and we know weather the file already exists or not

// If the file did exist
	if(fileExists)
	{
		printf("Found a file called '%s' in FAT records at '%d'\nChecling if it is open...\n", existingFATentry.filename, existingFATentry.address); 

//Look to see if the FILE is opened already. To do so, start at the FAT index and look for EOF. 
		for(i = existingFATentry.address; i<1000;i++)
			{
				fseek(FILES,i, 0);
				fgets(byte,2,FILES);		  		//Grab a byte
				if(byte[0] == '1')
					break;
//DEBUGGING
//printf("byte[%d]=%s\n",i,byte);			
				// During the scan, if we hit an empty spot or a buffer zone before the EOF pointer
				if(byte[0] == '0' || byte[0] == '9')
				{
					printf("Sorry, file '%s' is already opened! Try again later. \nOperation cancelled!\n...\n",newFATentry.filename);
					return -6;
				}		
			}
		// Also need to account for the fact that we reached end of disk... I don't know if this is a possible situation yet (haven't written the code yet)
		// But better have a safeguard
		if(i == 999 && byte[0] != 1)
			{
				printf("Sorry, file '%s' is already opened! Try again later. \nOperation cancelled!\n...\n",newFATentry.filename);
				return -6;
			}		

// If the existing file is not already opened, delete EOF pointer to notify others that it is now open

		// From the last scan, i already points to the EOF bit		
		fseek(FILES,i, 0);
		fprintf(FILES, "9");

		// Update file handle information. The file pointer will point to the start of the file, ready to overwrite, but we'll only do this if a WRITE command is issued
		FileHandles[INDEX][0] = existingFATentry.address;
		FileHandles[INDEX][1] = PIDRunning;	
		printf("Preparing to overwrite '%s' from HDD %d (Assigned file handle #%d)\n",newFATentry.filename, i, INDEX);
// We also set it up a write buffer to avoid having a second open file happen next door and mess exerything up
		// We use 9 as an indicator of buffer space; the buffer will extend from the existing file 
		makeBuffer(existingFATentry.address+1);
		return 0;
	}// End of of the file already existed



// If the file didn't exist
	// Scan the HDD for an empty location
	if(!fileExists)
	{
		rewind(FILES);	
		HDDpointer = -1;
		for(i=0;i<1000;i++)
		{
			fgets(byte,2,FILES);  		//Grab a byte
			if(byte[0] == '0')					// If the byte extracted was 0
			{			
				HDDpointer = i;
				printf("Found empty spot at HDD %d\n",HDDpointer);
				break;
			}
		}
	
	// If there is no more room in the HDD, 
		if (HDDpointer == -1)
		{
			printf("Process PID0x0%d terminate. No more disk space.\n\n",PIDRunning);
			return -1;   // This will send the process into the terminate Q, as per assignment requirements
		}



	// Then update FAT records either by filling an empty spot or appending
	// Otherwise, using FAT which is now pointing at the end of the FAT file

	rewind(FAT);
	countFATRecords = 0;
 	while (feof(FAT) == 0)
  	{
		memset(existingFATentry.filename, '\0', sizeof(existingFATentry.filename));		// Format buffer avoid accidental
		countFATRecords++;
    	fread(&existingFATentry, sizeof existingFATentry, 1, FAT);
		if(existingFATentry.address == -1)												// TAG for deleted FAT entries
			break;
	}
		// Line up the FAT pointer to the empty spot (or it points to the end of the FAT file)
		countFATRecords = --countFATRecords*24;
		fseek(FAT,countFATRecords, 0);


		newFATentry.address = HDDpointer;
		printf("Creating FAT record for '%s' at HDD %d\n", newFATentry.filename, HDDpointer);
		fwrite(&newFATentry, sizeof newFATentry, 1, FAT);
		FileHandles[INDEX][0] = newFATentry.address;
		FileHandles[INDEX][1] = PIDRunning;	
		printf("Assigned file handle #%d\n",INDEX);

	// We also set it up a write buffer to avoid having a second open file happen next door and mess exerything up
	// We use 9 as an indicator of buffer space; 
		makeBuffer(HDDpointer);
	}
	return 0;
}

/*******************************************************************************/
/*	Function FreeResourceIndex(int INDEX)
/*******************************************************************************/


void FreeResourceIndex(int INDEX)
{
	int i;
	char byte[2];

//FOR debugging
//for(i=0;i<5;i++)
//	printf("FileHandles[%d][0] = %d\tFileHandles[%d][1] = %d\n",i,FileHandles[i][1],i,FileHandles[i][1]);


	printf("Assigning EOF @ HDD pointer %d\n",FileHandles[INDEX][0]);		// Tell the user
	fseek(FILES,FileHandles[INDEX][0], 0);				// Go to closing point
	fprintf(FILES, "1");								// Mark EOF


	// Then delete the buffer
	fgets(byte,2,FILES);		  		//Grab a byte
	for(i=1; i<1000 && byte[0] == '9';i++)
	{
		fseek(FILES,(FileHandles[INDEX][0]+i), 0);
		fgets(byte,2,FILES);		  		//Grab a byte

//Debugging
//printf("Resetting byte[%d] = %s\n",(FileHandles[INDEX][0]+i),byte);
		if(byte[0] == '9')						// If the byte is empty
			{
				fseek(FILES,(FileHandles[INDEX][0]+i), 0);
				fprintf(FILES, "0");				// copy 9 to indicate that the space is reserved for the file handle buffer 
			}
	}
	FileHandles[INDEX][0] = -1;							// Free the ressource
	FileHandles[INDEX][1] = -1;							
	clearedHandles[INDEX] = 1;							// Tell the OS

//FOR debugging
//for(i=0;i<5;i++)
//	printf("FileHandles[%d][0] = %d\tFileHandles[%d][1] = %d\n",i,FileHandles[i][1],i,FileHandles[i][1]);
}


/*******************************************************************************/
/*	Function WriteToFile(char userInput[], int PIDRunning)
/*******************************************************************************/

int WriteToFile(char userInput[], int PIDRunning)
{
	int INDEX;
	int i,j,k;
	int ILLEGAL;
	char byte[2];
	char STRING[1000];
	int BUFFER;

// For debugging and understanding what is going on
	printf("EXECUTING '%s'\n",userInput);

// Extract Index 
    // skip white spaces after WRITE
    for(i=5;i<200 && userInput[i]!='\0' && userInput[i]==' ';i++);


	// Extract index #
	// Reading ASCII char as an integer: [ASCII]49 = [int]1 and [ASCII]57 = [int]9
	INDEX = 0;
	// We get kicked out of this loop when we hit ' ' or ','
	while(userInput[i]!= ',' && userInput[i]!= ' ')
		{
			INDEX = (INDEX*10);						// Shift old value << 1 dec. place
			INDEX = INDEX + (userInput[i]-48);    	// Add LS digit
			i++;
			if(userInput[i] == '\0')
			{	
				printf("\nInstruction '%s' ignored!\n(correct syntax: WRITE INDEX, STRING)\n", userInput);
				return -8;
			}
		}
	if(INDEX < 0 ||INDEX > 4)
	{
		printf("\nInstruction '%s' ignored!\n", userInput);
		printf("(Illegal INDEX pointer (%d) => must be between [0..4])\n\n",INDEX);
		return -5;
	}


// Check to see if process had the right to write
	if(FileHandles[INDEX][1] != PIDRunning)
	{
		if(FileHandles[INDEX][1] == -1)
			printf("INDEX '%d' needs to be opened before writing!\nInstruction ignored!\n\n",INDEX);
		else
			printf("INDEX '%d' is owned by PID0x0%d, which is NOT you (PID0x0%d)\nInstruction ignored!\n\n",INDEX,FileHandles[INDEX][1],PIDRunning);
		return -8;
	}

    // skip white spaces after INDEX,
	i++;
    for(;i<200 && userInput[i]!='\0' && userInput[i]==' ';i++);



// Extract STRING
	ILLEGAL = 0;
    for (j=0;j<999 && userInput[i]!='\0';i++)
	{	// ASCII 48 = 0, ASCII 57 = 9
		if(userInput[i]>47 && userInput[i]<58)
		{
			printf("ILLEGAL character => '%c' IGNORED!\n",userInput[i]);
			ILLEGAL = 1;
		}		
		else
        	STRING[j++]=userInput[i];   
	}

	STRING[j] = '\0';
	if(ILLEGAL)
		printf("STRING reduced to: '%s'\n",STRING);

// At this point, we have string, INDEX and we are allowed to write

// Check to see if we are overwriting
// When we opened the file, function makeBuffer transformed to 9 every 0 bits
// So if we don't point at a 9, we are overwriting, so delete old content
// The EOF pointer already got deleted by the OPEN function
	fseek(FILES,FileHandles[INDEX][0], 0);
	fgets(byte,2,FILES);		  		//Grab a byte	

//DEBUGGING
//printf("byte = %s\n",byte);
	if(byte[0] != '9')	
	{
		printf("Deleting previous file content!\n");
		for(i=0; i<1000; i++)
		{
			if(byte[0] == '9' || byte[0] == '0')
				break;
			// Set the byte to 9 if we are within the first 100 byte of the pointer (same as create buffer function)
			fseek(FILES,(FileHandles[INDEX][0] +i), 0);
			if(i<100)
				fprintf(FILES, "9");
			// Otherwise set it to 0 for now and call the buffer later if needed
			else
				fprintf(FILES, "0");
			// Then grab the next guy in line
			fgets(byte,2,FILES);		  		//Grab a byte	
		}
	}// End if


// Now weather this is a new file, or overwriting, we are either staring at a wide empty buffered space (byte = 9) or at a wall
// The algrithm for writing is write, increment, look 
	for(i=0;FileHandles[INDEX][0] < 1000 && STRING[i] != '\0'; FileHandles[INDEX][0]++, i++)
	{
//printf("FileHandles[%d][0] = %d\n",INDEX,FileHandles[INDEX][0]);
//printf("byte = %s\n",byte);
		// Check if we reached end of disk
		if(FileHandles[INDEX][0] == 999)
		{
			printf("Unexpected instruction termination! Disk is full.\n\n");
			FreeResourceIndex(INDEX);
			return -5;
		}
		// Write the next byte
		fseek(FILES,FileHandles[INDEX][0], 0);
		BUFFER = STRING[i];		
		fputc(BUFFER, FILES);
		// Extract the next guy
		fgets(byte,2,FILES);
//printf("byte = %s\n",byte);
		// If the nextguy is not buffer but is free, invoke make buffer from disk location
		if(byte[0] != '9')
		{
			if(byte[0] == '0')
				makeBuffer(FileHandles[INDEX][0]);
			// Otherwise, we are out of buffer and the next bit is taken. Invoke release resource
			else
			{
				printf("Unexpected instruction termination! Ran out of contigeous space to write\n\n");
				FreeResourceIndex(INDEX);
				return -5;
			}
		}// End if if the next byte is not 9

	} // End of for loop
// Now we reached the end of the string

	return 0;
}


/*******************************************************************************/
/*	Function closeINDEX(char userInput[], int PIDRunning)
/*******************************************************************************/

int closeINDEX(char userInput[], int PIDRunning)
{
	int INDEX,i;	
// For debugging and understanding what is going on
	printf("EXECUTING '%s'\n",userInput);

// Extract Index 
    // skip white spaces after CLOSE
    for(i=5;i<200 && userInput[i]!='\0' && userInput[i]==' ';i++);


	// Extract index #
	// Reading ASCII char as an integer: [ASCII]49 = [int]1 and [ASCII]57 = [int]9
	INDEX = 0;
	// We get kicked out of this loop when we hit NULL
	while(userInput[i])
		{
			INDEX = (INDEX*10);						// Shift old value << 1 dec. place
			INDEX = INDEX + (userInput[i]-48);    	// Add LS digit
			i++;
		}
	if(INDEX < 0 ||INDEX > 4)
	{
		printf("\nInstruction '%s' ignored!\n", userInput);
		printf("(Illegal INDEX pointer (%d) => must be between [0..4])\n\n",INDEX);
		return -5;
	}


// Check to see if process owns the INDEX
	if(FileHandles[INDEX][1] != PIDRunning)
	{
		if(FileHandles[INDEX][1] == -1)
			printf("INDEX '%d' is not in use at the moment...\nInstruction ignored!\n\n",INDEX);
		else
			printf("INDEX '%d' is owned by PID0x0%d, which is NOT you (PID0x0%d)\nInstruction ignored!\n\n",INDEX,FileHandles[INDEX][1],PIDRunning);
		return -8;
	}

// Then call the function that clears the buffer, assigns EOF and releases the FileHandle
	FreeResourceIndex(INDEX);
	return 0;
}


/*******************************************************************************/
/*	Function deleteFile(char userInput[], int PIDRunning)
/*******************************************************************************/

deleteFile(char userInput[], int PIDRunning)
{
	int INDEX,i,j;
	char FILENAME[20];
	char byte[2];
	struct OSFAT existingFATentry, emptyFAT;
	int fileExists = 0;
	int countFATRecords=0;

// For debugging and understanding what is going on
	printf("EXECUTING '%s'\n",userInput);

// Extract Index 
    // skip white spaces after DELETE
    for(i=6;i<200 && userInput[i]!='\0' && userInput[i]==' ';i++);


// Extract FILENAME
    for (j=0;j<19 && userInput[i]!='\0';i++)
        FILENAME[j++]=userInput[i];   


	FILENAME[j] = '\0';

	// Display error if filename too long
	if(userInput[i]!='\0')
	{
		printf("\nInstruction '%s' ignored!\n", userInput);
		printf("(Filename restricted to 20 characters!)\n\n");
		return -5;
	}


	// Look up the filename and HDD location from the FAT too make sure the file exists
	rewind(FAT);
	// fileExists initialized at 0

	printf("\nLooking up FAT records for '%s'...\n",FILENAME);
	// Scan FAT table until EOF or until found existing file name
 	while (feof(FAT) == 0 && fileExists == 0)
  	{
		memset(existingFATentry.filename, '\0', sizeof(existingFATentry.filename));		// Format buffer avoid accidental
		countFATRecords++;
    	if(!(fread(&existingFATentry, sizeof existingFATentry, 1, FAT)))
			{
				printf("Reached end for FAT\n");	// Read next record on OSFAT.txt file
				break;
			}
		else
			printf("-rwxrw-rw-\tPID0x0%d\t  %d\t%s\n",PIDRunning,existingFATentry.address, existingFATentry.filename);
	
		fileExists = 1;	                						// Assume File exists until proven wrong
		for(i = 0; existingFATentry.filename[i] != '\0'; i++)	// Check if File exists
    		{
    			if(existingFATentry.filename[i] != FILENAME[i])	        // Character by character analysis
    				{
                        fileExists=0;			                // Any divergence between the 2 values
						break;                                  // kicks the porogram out of this loop
                    }// End of looking filename
		    }// End for(scanning filename)
	}// Reached EOF without finding a matching filename name
	printf("\n");

// If the file doesn't exist, tell the user and ignore command
	if(!fileExists)
	{
		printf("Unable to process request! File '%s' not found in FAT table!\n\n",FILENAME);
		return -5;
	}




// If the file did exist, make sure it's not open
	if(fileExists)
	{
		printf("Found a file called '%s' in FAT records at '%d'\nChecling if it is open...\n", existingFATentry.filename, existingFATentry.address); 

//Look to see if the FILE is opened already. To do so, start at the FAT index and look for EOF. 
		for(i = existingFATentry.address; i<1000;i++)
			{
				fseek(FILES,i, 0);
				fgets(byte,2,FILES);		  		//Grab a byte
				if(byte[0] == '1')
					break;
//DEBUGGING
//printf("byte[%d]=%s\n",i,byte);			
				// During the scan, if we hit an empty spot or a buffer zone before the EOF pointer
				if(byte[0] == '0' || byte[0] == '9')
				{
					printf("Sorry, file '%s' is already opened! Try again later. \nOperation cancelled!\n...\n",FILENAME);
					return -6;
				}		
			}
		// Also need to account for the fact that we reached end of disk... I don't know if this is a possible situation yet (haven't written the code yet)
		// But better have a safeguard
		if(i == 999 && byte[0] != 1)
			{
				printf("Sorry, file '%s' is already opened! Try again later. \nOperation cancelled!\n...\n",FILENAME);
				return -6;
			}		

// Here, we know for sure that the file exists and that it is not open. existingFATentry.address points to the start of the file
// We also know that at the very least, all there is to the file is an EOF pointer, otherwise, stuff then EOF
		byte[0]= '0';
		printf("Deleting previous file content!\n");
		for(i=0; i<1000 && byte[0] != '1'; i++)
		{

			fseek(FILES,(existingFATentry.address +i), 0);			// Go to start of file + i offset
			fprintf(FILES, "0");									// Delete byte
			fgets(byte,2,FILES);		  							//Grab the next byte	
//DEBUGGING
//printf("byte[%d]=%s\n",i,byte);	
			if(byte[0] == '1')										// If we reached EOF
			{
//printf("LAST byte[%d]=%s\n",i,byte);	
//THE COMPILER IGNORES THIS L fprint FOR SOME REASON!
				fseek(FILES,(existingFATentry.address +i+1), 0);	// Get the pointer there
				fprintf(FILES, "HERE");								// Delete EOF pointer
				break;												// And done deleting file!
			}
		}// END for loop

// Now we need to delete old fat record. countFATRecords counted the records read until we found the one to delete (every record is 24 bytes)
// So the offset we need is one count before countFATRecords, since the 1st record is at offset 0
		printf("Deleting FAT record for '%s'\n", FILENAME);
		strcpy(emptyFAT.filename,"EMPTY_RECORD");
		emptyFAT.address = -1;		
		countFATRecords = --countFATRecords*24;
		fseek(FAT,countFATRecords, 0);
		fwrite(&emptyFAT, sizeof emptyFAT, 1, FAT);

//DEBUGGING
	rewind(FAT);
	printf("\nUpdated FAT record\n");
 	while (feof(FAT) == 0)
  	{
		memset(existingFATentry.filename, '\0', sizeof(existingFATentry.filename));		// Format buffer avoid accidental
    	if(!(fread(&existingFATentry, sizeof existingFATentry, 1, FAT)))
			{
				printf("Reached end for FAT\n");	// Read next record on OSFAT.txt file
				break;
			}
		else
			printf("-rwxrw-rw-\tPID0x0%d\t  %d\t%s\n",PIDRunning,existingFATentry.address, existingFATentry.filename);
	}

	}// End of if file exists






	return 0;
}


/*******************************************************************************/
/*	Function  main (int argc, char *argv[])
/*******************************************************************************/

int main (int argc, char *argv[])
{

	int returnCode = 1;
    int progEnd, i;
	int INDEX;
    char instruction[100];
	int PIDloading;
	int RAMAddress;	
	int DONE;
	int pageNumber;
	int numRunning;
	int numWaiting;

/****************Step 1: Check for switches***********************************/

	if(--argc)
		if(switches(argc, argv))
			return -1;
	printf("Because of the FRAMESIZE\n", FRAMESIZE);
	printf("Only %d script(s) will be able to run simultaneously\n",concurrent);
	printf("(Running %d scripts will greatly decrease the performance of the machine!)\n\n",concurrent);
	

/****************Step 2: Initialize RAM and File system ***************************************/
	if(formatRAM())puts("WARNING!\n\nThe OS detected an error while formating RAM");


// Create and format filesystem
	if(createFS() == -1)printf("Unexpected error creating file system... OS might not work properly\n");


/****************Step 3: Initialize Run-Time Environment************************/

    for(i=0; i<10; i++)                             // Initialization of Queues
        {
            readyQ[i] = -1;                         // -1 is used to indicate an empty spot (since all PIDs are >= 0
            terminateQ[i] = -1;
		}
	for(i=0; i<50; i++)                             // Initialization of Loading Q
	{
		loadingQ[i] = -1;
	}

	for(INDEX=0; INDEX<5; INDEX++)
	{
		for(i=0; i<50; i++)                             // Initialization of Waiting Q for each file index
			waitingQ[INDEX][i] = -1;

		waitingHead[INDEX] = 0;							        // Initialization of Q pointers
    	waitingTail[INDEX] = 0;  

	}


	readyHead = 0;							        // Initialization of Q pointers
    readyTail = 0;  
	loadingHead = 0;							    // Initialization of Q pointers
    loadingTail = 0;                                               
	int PIDRunning;								    // Variable to Identify Running PCB
// initiate cleared handles flag
	for(i=0; i<5; i++)
		clearedHandles[i] = 0;

/*******************************************************************************/
/*	3. Initialization of Virtual CPU with a DATA structure
/*******************************************************************************/

	struct cpu CPU;
	struct cpu *CPUptr;
	CPUptr = &CPU;	

	CPU.quanta = quanta; 							    // The value of quanta can be set by user
	CPU.returnCode = 0;
	CPU.pc = 0;
	memset(CPU.memory, '\0', sizeof(CPU.memory));


/****************Step 4: Run-Time Environment**********************************/


	mkdir("BACKINGSTORE", 0777);// Cuz we'll need it

// Format file handles
	for(i=0; i<5;i++)
	{
		FileHandles[i][0] = -1;
		FileHandles[i][1] = -1;
	}	

while(!login());						            	        //Can only pass this step if valid credentials
   {
   printf("Loading shell\n\n");

   while (returnCode != -1)
     {  
        printf("\n");                                           // Just for esthetics
	    returnCode = shell();	 					            // returns -1 when user inputs EXIT or LOGOUT

// Figure out who gets to run next 

// First overwrite whe was running last
	PIDRunning = -1;
// If a file handle got cleared, got that pore process out of his misery in the waiting Q!
		for(i=0; i<5; i++)
		{	// a 1 flag = handle cleared, o flag = no activity			
//printf("ClearedHandles[%d] = %d\n",i,clearedHandles[i]);
			if(clearedHandles[i])
			{	// The waiting Q is segmented by INDEX
				PIDRunning = PopWaitingQ(i);			// Extract waiting processes
				clearedHandles[i] = 0;					// Clear the flag
				break;
			}
		}

// NEXT 20 COMMENTED OUT LINES FOR DEBUGGING Qs
/*
int INDEX;
for(INDEX = 0; INDEX<5;INDEX++)
{
	for(i=0; i<20;i++)
		printf("waitingQ[%d][%d] = %d\n",INDEX, i, waitingQ[INDEX][i]);
	printf("waiting[%d]H = %d, waiting[%d]T = %d\n\n",INDEX, waitingHead[INDEX],INDEX, waitingTail[INDEX]);
} 
// For debugging Qs:
for(i=0;i<10;i++)
	printf("Loading[%d] = %d   Ready[%d] = %d, terminate[%d] = %d\n",i,loadingQ[i],i,readyQ[i],i,terminateQ[i]);
printf("Ready H=%d T=%d \nLoading H=%d T=%d\n",readyHead,readyTail,loadingHead, loadingTail);
*/
//DONE DEBUGGING


// The pop operation returns the PID of the next guy on the appropriate INDEX wait Q or -1 if there was no one on the Q
//printf("PIDRunning = %d\n",PIDRunning);

// Otherwise, look for the next guy in the readyQ  	    
		if(PIDRunning == -1)
			PIDRunning = PopReadyQueue();                           // Pops next PCB from ready Q
// If there is no one to run, reset Q pointers
//printf("THEN PIDRunning = %d\n",PIDRunning);


// Run next guy on the running Q
        if(PIDRunning != -1)
            {
                printf("CPU is running PID 0x0%d (called '%s')\n", PIDRunning, PCB[PIDRunning].filename); 
	            TaskSwitchIn(CPUptr, PIDRunning);	    		// Sets PC, pops variable from 'stack' to 'CPU registers'
	            CPU.quanta = quanta; 					        // Recursive step because this is done in TaskSwitchIn()...
				DONE = 0;										// Not done

                while(CPU.quanta > 0)                           // while Quanta left	       
              	   {

// If the instruction is in RAM
					pageNumber = CPU.pc/4;												// First compute which page contains the instruction
					if((PCB[PIDRunning].pageTable[pageNumber].inRAM))					// Then look up the page table
					{
						RAMAddress = PCB[PIDRunning].pageTable[pageNumber].baseAddress;	// Look up base address for the page
						RAMAddress += CPU.pc%4;											// Calculate instruction address offset

// FOR DEBUGGING
// printf("PC = %d  Page#= %d RAM address=%d\n",CPU.pc,pageNumber,RAMAddress);


// Execute the instruction
                      	CPU.returnCode = interpreter(FREERAM[RAMAddress], PIDRunning);  // Send command and PCB memory address		
              		  	CPU.pc++;						                        		// Increment CPU program counter		
                    	CPU.quanta --;
					 }// End of if the instruction was in RAM


// Else, if the current instruction of the PCB[PIDrunning] is not in RAM, call the memory manager to store it there
					else
					{
						printf("Instruction %d is not in RAM...\nFetching it in the BS and putting PID0x0%d to sleep.\n\n", CPU.pc, PIDRunning);
						PCB[PIDRunning].current = CPU.pc;		// Since the memMGR relies on this info to build page table
						memManager(PIDRunning);					// Page reference for next frame to be loaded in RAM
						CPU.quanta = 0;							// Kicks us out of CPU until next time
					}

// Then, evaluate weather an EXIT command got processes, or we are at the end of the script
                     if(CPU.returnCode == -1 || (CPU.pc == PCB[PIDRunning].length && CPU.returnCode != -2 ))       // Condition that program is done running
	               	   {    
                     	printf("\n\n");
                    	CPU.quanta = 0;				       		                        // Break out of CPU quanta loop
						DONE = 1;														// Records the fact that we are done executing                    	
						}

// If we got a wait for file handle return code from the interpreter
					 if(CPU.returnCode == -2)
					 {
                    	CPU.quanta = 0;	       		                        // Break out of CPU quanta loop
						TaskSwitchOut(CPUptr, PIDRunning);                  // Save runtinme environment in PCB
// INDEXwait is like a register containing the file Handle index that PID Running is waiting for 
	                	if(PushWaitingQ(PIDRunning, INDEXwait) != -1)
							printf("PID0x0%d pushed in the waiting Q\n",PIDRunning);
						else
							printf("waiting Q is full. Instruction ignored\n");	
						}// End of returncode == -2
					}// End of CPU Quanta

// If the program was done executing
				if(DONE)
                	PushTerminate(PIDRunning);	       				                // Pushes the program into the terminate Queue					

// Otherwise, if there was still some instructions to execute
				if(!DONE && CPU.returnCode != -2)
					{
						TaskSwitchOut(CPUptr, PIDRunning);                  // Save runtinme environment in PCB 
						PushReadyQueue(PIDRunning);                     // Pushes current program into the Ready Q
					}
            }// End if(PIDRunning != -1)

// More OS Overhead
        TerminateProcesses();

// Count how many processes are running
	numRunning = 0;
	for(i=0; i<10;i++)
		if(readyQ[i] != -1)									
			numRunning++;

	for(INDEX=0; INDEX < 5; INDEX++)	
	{	for(i=0; i<50;i++)
			if(waitingQ[INDEX][i] != -1)									
				numRunning++;
	}
// Count how many processes are waiting to load (just for fun)
	numWaiting = 0;
	for(i=0; i<50;i++)
		if(loadingQ[i] != -1)									
			numWaiting++;
	
// If there are less processes running that what is allowed based on framesize
	if(numRunning < concurrent)
	{
		PIDloading = PopLoadingQ();								// Pop the loading Q to see if we need to load a program

// Some OS Overhead: Load next guy on the loading Q (if any)
		if(PIDloading != -1)
			{
				memManager(PIDloading);				// Loads PCB's first page int RAM, then pushes him into readyQ by the memMGR
				PushReadyQueue(PIDloading);                     // Pushes current program into the Ready Q
			}
	}

// For fun
		printf("There is/are %d Processe(s) running at the moment\n(%d Allowed in total, %d Waiting)\n",numRunning,concurrent, numWaiting);


    }// End while LOGGED into the SHELL
	puts("Logging out of the shell and exiting OS\n");


// Close the FS files
	
	fclose(FAT);
	fclose(FILES);



}//End of while loop initiating CPU environment... Equivalent of closing computer
// The code can't get here.
return 0;
}// End of main()


// Just for fun, because programing gets boring after a while...
void logo(void)
{
puts("\n"
"           (((\n"
"           {  '>\n"
"           / ,)\n"
"          /_/ /\n"
"==========//=n===========\n"
"         /\n\n"
"    WELCOME TO BIRDX\n\n");
}

